/*
 * First attempt at an ICFP 2002 robot implemented in Java
 */

import java.lang.*;
import java.util.*;
import java.io.*;
import java.net.*;

public class ICFPContest {

	// connnection and global data
	public static String host;
	public static int port;
	public static  java.net.Socket sock;
	private static BufferedReader input;
	private static PrintWriter output;
	public static int verbosity = 0;
	public static String logfile;

	// map data
	public static int mapWidth;
	public static int mapHeight;
	public static char mapData[][];

	// global constants
	// public static Integer Player.robotID;
	// public static int carryingCapacity;
	// public static int startingBalance;

	// my data
	// public static int balance;
	// public static Coords currLoc;

	// robot and package data
	public static int robotCount = 1;
	public static Hashtable robotLocs;
	public static Hashtable robotInv;
	public static Hashtable packages;
	public static Hashtable packLocations;
	public static Hashtable herePackages;


	public static void main( String argv[] ) {
		try {
			host = argv[0];
			port = Integer.parseInt( argv[1] );
	
			robotLocs = new Hashtable();
			robotInv = new Hashtable();
			packages = new Hashtable();
			packLocations = new Hashtable();
			herePackages = new Hashtable();
	
			for(int i=2; i < argv.length; i++) {
				if ( argv[i].equals("--verbose") ) 
				{
					try {
						verbosity = Integer.parseInt ( argv[++i] );
					}
				catch (Throwable e) {
						verbosity = 1;
					}
				} else if ( argv[i].equals("--logfile") ) {
					logfile = argv[++i];
				} else {
					throw new IllegalArgumentException("Invalid Parameter: " + argv[i]);
				}
	
			}
			init();


		}
		catch (UnknownHostException e) {
			log("Can't resolve host " + argv[0],-1);
			System.exit(1);
		}
		catch (IllegalArgumentException e) {
			usage();
			System.exit(2);
		}
		catch (IOException e) {
			log("Premature server disconnect",-1);
			e.printStackTrace();
			System.exit(3);
		}


		try {
			run();
		}
		catch(Throwable e) {
			log("Died because "+e.toString(), 2);
			if ( verbosity > 0 )
			{
				e.printStackTrace();
			}
		}
		finally {
			log("Final score: "+Player.score, 0);
			log("remaining money: "+Player.balance, 1);
		}

	}

	public static void usage() {
		System.err.println("ICFPRobot host port [--verbose n] [--logfile file]");
	}

	private static void init ( ) throws IOException {

		// Open the connection 
		sock = new java.net.Socket( host, port );
		
		log ( "Connected to server " + host + " on port " + Integer.toString(port) , 1);

		// Set up input and putput streams

		input = new BufferedReader( new InputStreamReader ( sock.getInputStream() ) );
		output = new PrintWriter( sock.getOutputStream(), /* autoflush */ true );

		// Send client type:
		
		sendLine("Player");

		// Read in Map
		
		int values[] = readIntegers();
		mapWidth = values[0]; 
		mapHeight = values[1];

		mapData = new char[mapWidth+2][mapHeight+2];

		for(int x=0; x <= mapWidth+1; x++) {
			mapData[x][0] = mapData[x][mapHeight+1] = '#';  // top/bottom
		}

		for(int y=1; y <= mapHeight; y++) {
			char line[] = readLine().toCharArray();
			mapData[0][y] = mapData[mapWidth+1][y] = '#'; // left/right
			for(int x=1; x <= mapWidth; x++) {
				mapData[x][y] = line[x-1];
			}
		}

		// Read User data
		values = readIntegers();
		Player.robotID = new Integer(values[0]);
		Player.carryingCapacity = Player.capacity = values[1];
		Player.startingBalance = Player.balance = values[2];

		// Initialize callbacks here


	}

	private static String readLine() throws IOException {
		String line = input.readLine();
		log("SERVER " + line, 2);

		return line;
	}

	private static void sendLine(String line) throws IOException {
		log("CLIENT " + line, 2);
		output.println(line);
	}

	private static int[] readIntegers () throws IOException {

		String tokens[] = readTokens();

		int values[] = new int[tokens.length];

		for (int i=0; i<tokens.length; i++) {
			try { values[i] = Integer.parseInt( tokens[i] ) ; }
			catch (NumberFormatException e) {
				throw new IOException( "Mangled number from server: " + tokens[i] );
			}
		}

		return values;

	}

	private static String[] readTokens() throws IOException {
		StringTokenizer tokens = new StringTokenizer( readLine() );
		
		int numTokens = tokens.countTokens();
		String list[] = new String[numTokens];

		for (int i=0; i < numTokens; i++) {
			list[i] = tokens.nextToken();
		}

		return list;
	}

	public static void run() throws IOException {
		while (true) {
			readStatus();

			readPackageLine();

			if ( verbosity > 2 ) {
				printMap(); 
			}
	
			sendMove();
		}

	}

	private static void sendMove() throws IOException {
		// really stupid, but good for testing

		Move nextMove = Player.getMove();

		sendLine( nextMove.toString() );

	}

	private static void readStatus() throws IOException {
		Integer currentRobot = new Integer(-1); // because it's used as an index to a Hashtable
		int eventCount = 0;

		robotCount = 0;

		String lastEvent;
		String currToken = "";

		Vector seen = new Vector();

		String tokens[] = readTokens();
		
		try {
			for ( int index = 0; index < tokens.length; index++) {
				currToken = tokens[index];
				if ( currToken.startsWith("#") ) {
					currentRobot = new Integer( currToken.substring(1) );
					robotCount++;
					eventCount = 0;
					log("Switching to robot "+currToken,5);
					if ( currentRobot.equals(Player.robotID) ) {
						log( "  That's ME!", 5 );
					}

					seen.addElement( currentRobot );

				} else if ( currToken.equals("N") || currToken.equals("S") || currToken.equals("W") || currToken.equals("E") ) {
					robotMove(currentRobot, currToken);
					if(++eventCount > 1)  {
						// callback multi-event
					}
					lastEvent = currToken;
					log("   it moved " + currToken,5);
				} else if ( currToken.equals("P") ) {
					robotPickup( currentRobot, new Integer(tokens[++index]) );
					if(++eventCount > 1) {
						// callback mulri-event
					}
					log("   it picked up package #" + tokens[index], 5);
				} else if ( currToken.equals("D") ) {
					robotDrop( currentRobot, new Integer(tokens[++index]) );
					if(++eventCount > 1) {
						// callback multi-event
					}
					log("   it dropped package #" + tokens[index], 5);
				} else if ( currToken.equals("X") ) {
					// I can assume that tokens[index+2].equals("Y")
					robotSetPos( 	currentRobot,
									Integer.parseInt( tokens[index+1] ), 
									Integer.parseInt( tokens[index+3] ) );
					log ("  moved it to X=" + tokens[index+1] + " Y=" + tokens[index+3], 5);
					index += 3;  // 3 because one will be added by the loop.
				} else {
					log("Invalid status update: " + currToken,3);
				}

			}
		} // try 
		catch (NumberFormatException e) {
			throw new IllegalArgumentException( "Invalid number from server: " + currToken);
		}

		// Check to see if anybody didnt' report in.
		Enumeration robotIDs = robotLocs.keys();
		while( robotIDs.hasMoreElements() ) {
			Integer id = (Integer) robotIDs.nextElement();
			if ( ! seen.contains( id ) ) {
				log("Robot #" + id.toString() + " is missing, assumed dead", 2);
				robotDeath( id );
			}
		}

	}

	private static void readPackageLine() throws IOException, IllegalArgumentException {
		herePackages = new Hashtable();

		int packInfo[] = readIntegers();

		if ( packInfo.length > 0 ) {
			log("Found "+(packInfo.length/4)+" packages here",5);
			// There are packages here, make a list.
			for(int i=0; i < packInfo.length; i += 4)
			{
					Integer packID = new Integer( packInfo[i] );
					Coords dest = new Coords( packInfo[i+1], packInfo[i+2] );
					ICFPPack pack = new ICFPPack( packID, new Coords(Player.currLoc), dest, packInfo[i+3] );

					log("   ID: " + packID + " Dest: "+dest.toString()+" Weight: " + packInfo[i+3], 5);
					herePackages.put(packID, pack);

					packages.put(packID, pack);
			}

			packLocations.put(Player.currLoc,herePackages);
		} else { 
			// otherwise delete location information for packs we thought were here
			if ( packLocations.containsKey(Player.currLoc) ) {
				Hashtable oldPackages = (Hashtable) packLocations.get(Player.currLoc);
				packLocations.remove(Player.currLoc);
				Enumeration oldPackEnum = oldPackages.elements();
				while ( oldPackEnum.hasMoreElements() ) {
					ICFPPack pack = (ICFPPack) oldPackEnum.nextElement();
					log("Deleating previous locaiton information for pack #"+pack.getID(), 5);
					pack.clearLoc();
				}
			} 
		}
	}

	private static void robotMove(Integer robotID, String dir ) {
		int newX;
		int newY;

		Coords currPos = (Coords) robotLocs.get( robotID );

		currPos.move ( dir );

		char newMapElement = mapData[currPos.x][currPos.y];
		if ( currPos.x < 1 || currPos.y < 1 || currPos.x > mapWidth || currPos.y > mapHeight ) {
			log("Robot " + robotID.toString() + "Moved off the map!!!!",0);
		} else {
			switch ( newMapElement ) {
				case '~':
					// Robot drowns
					log("Robot #" + robotID.toString() + " walks (or was pushed) into the water and drowns!",2);
					robotDeath( robotID );
					break;
				case '#':
					log("Robot " + robotID.toString() + "Moved into WALL!!!",-1);
					break;
				case '@':
					// base arrival notification callback?
					break;
			}

			if ( robotID.equals( Player.robotID ) ) {
				Player.currLoc = new Coords( currPos );
			}
		}

	}

	private static void robotDeath ( Integer robotID ) {

		robotLocs.remove(robotID);
		if ( robotInv.containsKey( robotID ) ) {
			Hashtable carrying = (Hashtable) robotInv.get(robotID);
			robotInv.remove(robotID);
			Enumeration invPacks = carrying.keys();
			while ( invPacks.hasMoreElements() ) {
				packages.remove( invPacks.nextElement() );
			}
		}
	}

	private static void robotSetPos (Integer robotID, int x, int y) {
		Coords pos;
		
		if ( robotLocs.containsKey( robotID ) ) {
			pos = (Coords) robotLocs.get( robotID );
			pos.setPosition( x, y );
		} else {
			robotLocs.put( robotID, pos = new Coords(x, y) );
		}

		if ( robotID.equals(Player.robotID) ) {
			Player.currLoc = pos;
		}
	}

	private static void robotPickup (Integer robotID, Integer packID) {
		Hashtable inv;

		if ( (inv = (Hashtable) robotInv.get( robotID ) ) == null ) {
			robotInv.put( robotID, inv = new Hashtable() );
		}

		Coords pos = (Coords) robotLocs.get( robotID );
		ICFPPack pack;

		if ( (pack = (ICFPPack) packages.get( packID ) ) == null ) {
			packages.put( packID, pack = new ICFPPack( packID ) );
		}

		pack.clearLoc();
		pack.setCarry(true);

		inv.put( packID , pack );

		if ( robotID.equals( Player.robotID ) ) {
			// Player.?lastTurnPickup = true; // hackish
			Player.capacity -= pack.getWeight();
			if( Player.capacity < 0 ) {
				// Can't Happen!!!!
				log("You picked up MORE than you can carry!!!",-1);
			}
			log("You picked up package #"+packID+" weighing "+pack.getWeight()+" leaving capacity of "+Player.capacity, 3);
		}

		if ( packLocations.containsKey( pos ) ) {
			Hashtable posInv = (Hashtable) packLocations.get( pos );
			posInv.remove( packID );
		}

	}

	private static void robotDrop (Integer robotID, Integer packID) {
		ICFPPack pack;
		Coords robotPos = (Coords) robotLocs.get( robotID );;
		Coords packPos = new Coords( robotPos );
		Hashtable inv = null;

		if ( (inv = (Hashtable) robotInv.get( robotID ) ) == null ) {
			robotInv.put( robotID, inv = new Hashtable() );
		}


		if ( inv.containsKey( packID ) )
		{
			inv.remove( packID );
			if ( robotID.equals( Player.robotID ) )
			{
				pack = (ICFPPack) packages.get( packID );
				Player.capacity += pack.getWeight();
				log(" by droping package #"+packID+" player increases capacity to "+Player.capacity, 3);
			}
		} else {
			log("Robot droped a package it wasn't carrying!",3);
		} 

		if ( packages.containsKey( packID ) ) {
			pack = (ICFPPack) packages.get( packID );
		} else {
			packages.put( packID, pack = new ICFPPack( packID ) );
		}

		if ( pack.destValid() && pack.getDest().equals(robotPos) ) {
			packages.remove( packID );
			log("Robot " + robotID.toString() + " delivered package " + packID.toString(), 2);
			if ( robotID.equals( Player.robotID ) ) {
				Player.score += pack.getWeight();
				// Player.balance += pack.getWeight();
				log(" you scored " + pack.getWeight() + " for delivering pack #"+packID+" which brings your total to " + Player.score, 1);
			}
			
		} else {
			pack.setLoc( new Coords( robotPos ) , false);
			pack.setCarry(false);
			// Do we notifify about possibles?
		}

	}

	protected static void log (String message, int minLogLevel) {
		if ( verbosity >= minLogLevel ) {
			PrintWriter logOutstream = null;
			boolean file = false;

			if ( logfile != null && logfile.length() > 0 ) {
				try {
					logOutstream = new PrintWriter( new FileWriter( logfile ) );
					file = true;
				}
				catch (Exception e) {
				}
			}
			if ( logOutstream == null ) {
				logOutstream = new PrintWriter( System.err );
			}

			logOutstream.println(message);

			logOutstream.flush();

			if ( file == true ) {
				logOutstream.close();
			}

		}
	}

	protected static void printMap() {
		
		char map[][] = new char[mapWidth+2][mapHeight+2];

		PrintWriter logOutstream = null;
		boolean file = false;

		if ( logfile != null && logfile.length() > 0 ) {
			try {
				logOutstream = new PrintWriter( new FileWriter( logfile ) );
				file = true;
			}
			catch (Exception e) {
			}
		}
		if ( logOutstream == null ) {
			logOutstream = new PrintWriter( System.err );
		}

		for (int y=0; y < mapHeight+2; y++) {
			for(int x=0; x < mapWidth + 2; x++) {
				Coords pos = new Coords(x, y);

				if ( packLocations.containsKey( pos ) ) {
					map[x][y] = '$';
				} else {
					map[x][y] = mapData[x][y];
				}
			}
		}

		Enumeration botIDs = robotLocs.keys();
		while ( botIDs.hasMoreElements() ) {
			Integer botID = (Integer) botIDs.nextElement();
			Coords botPos = (Coords) robotLocs.get( botID );

			if ( botID.equals(Player.robotID) ) {
				map[botPos.x][botPos.y] = '!';
			} else {
				map[botPos.x][botPos.y] = Character.forDigit( botID.intValue() % 36, 36);
			}
		}
		
		for ( int y=mapHeight+1; y >= 0; y-- ) {
			char line[] = new char[mapWidth+2];
			for( int x=0; x < mapWidth+2; x++ ) {
				line[x] = map[x][y];
			}
			log(new String(line),2);
		}
	}

	// public accessor methods.

	/* public static Coords getCurrentLocation() {
		return Player.currLoc;
	} */

	/* public static int getCarryingCapacity() {
		return Player.carryingCapacity;
	} */

	/* public static int getMoneyBalance() {
		return Player.balance;
	} */

	/* public static int getStartingBalance() {
		return Player.startingBalance;
	} */

	/* public static Integer getMyRobotID() {
		return Player.robotID;
	} */

	public static ICFPPack getPackageInfo ( Integer packID ) {
		return (ICFPPack) packages.get( packID );
	}

	public static Enumeration getKnownPackIDs() {
		return packages.keys();
	}

	 public static Enumeration getKnownPackages() {
		return packages.elements();
	}

	public static Hashtable getAllKnownPackages() {
		return packages;
	} 

	public static Hashtable getPackagesAt( Coords there ) {
		return (Hashtable) packLocations.get( there );
	} 

	public static Hashtable getPackagesHere() {
		return herePackages;
	}

	public static Hashtable getRobotInventory ( Integer robotID ) {
		return (Hashtable) robotInv.get( robotID );
	}

	/* public static Hashtable myInventory() {
		return getRobotInventory( Player.robotID );
	} */
	
}



